@@ -1,10 +1,11 @@
 __author__ = "SutandoTsukai181"
 __copyright__ = "Copyright 2021, SutandoTsukai181"
 __license__ = "MIT"
-__version__ = "1.3.1"
+__version__ = "1.4.0"
 
 import struct
 from contextlib import contextmanager
+from enum import Flag, IntEnum
 from typing import Tuple, Union
 
 FMT = dict()
@@ -18,21 +19,63 @@ for c in ["q", "Q"]:
     FMT[c] = 8
 
 
+class Endian(Flag):
+    LITTLE = False
+    BIG = True
+
+
+class Whence(IntEnum):
+    BEGIN = 0
+    CUR = 1
+    END = 2
+
+
+class BrStruct:
+    """Base class for objects passed to BinaryReader's `read_struct` and `write_struct` methods.\n
+    Any type passed to `read_struct` and any object passed to `write_struct` must inherit from this class.\n
+    Override `__br_read__` and `__br_write__` methods from this class to set up BinaryReader to read your classes.\n"""
+
+    def __init__(self) -> None:
+        """If this class will be used with BinaryReader's `read_struct` method, then this method MUST receive zero arguments after `self`.\n
+        """
+        pass
+
+    def __br_read__(self, br: 'BinaryReader', *args) -> None:
+        """Called once when `BinaryReader.read_struct` is called on this class.\n
+        This method must accept at least 1 parameter (other than `self`).\n
+        The first parameter will be the BinaryReader instance which `read_struct` was called from.
+        This parameter can be used to `read` the attributes of object.\n
+        This method can take any number of parameters after the required first parameter.
+        The additional arguments corresponding to these parameters should be passed to `BinaryReader.read_struct` after the `count` argument.\n
+        """
+        pass
+
+    def __br_write__(self, br: 'BinaryReader', *args) -> None:
+        """Called once when `BinaryReader.write_struct` is called on an instance of this class.\n
+        This method must accept at least 1 parameter (other than `self`).\n
+        The first parameter will be the BinaryReader instance which `write_struct` was called from.
+        This parameter can be used to `write` the attributes of object.\n
+        This method can take any number of parameters after the required first parameter.
+        The additional arguments corresponding to these parameters should be passed to `BinaryReader.write_struct` after the `value` argument.\n
+        """
+        pass
+
+
 class BinaryReader:
     """A buffer reader/writer containing a mutable bytearray.\n
     Allows reading and writing various data types, while advancing the position of the buffer on each operation."""
     __buf: bytearray
     __idx: int
-    __big_end: bool
+    __endianness: Endian
     __encoding: str
 
-    def __init__(self, buffer=bytearray(), big_endian=False, encoding='utf-8'):
+    def __init__(self, buffer: bytearray = bytearray(), endianness: Endian = Endian.LITTLE, encoding='utf-8'):
         """Constructs a BinaryReader with the given buffer, endianness, and encoding and sets its position to 0.\n
         If buffer is not given, a new bytearray() is created. If endianness is not given, it is set to little endian.\n
         Default encoding is UTF-8. Will throw an exception if encoding is unknown.
         """
         self.__buf = bytearray(buffer)
-        self.__big_end = big_endian
+        self.__endianness = endianness
         self.__idx = 0
         self.set_encoding(encoding)
 
@@ -46,6 +89,17 @@ class BinaryReader:
         """Returns the current position in the buffer."""
         return self.__idx
 
+    def __past_eof(self, index: int) -> bool:
+        return index > self.size()
+
+    def past_eof(self) -> bool:
+        """Returns True if the current position is after the end of file."""
+        return self.__past_eof(self.pos())
+
+    def eof(self) -> bool:
+        """Returns True if the current position is at/after the end of file."""
+        return self.__past_eof(self.pos() + 1)
+
     def size(self) -> int:
         """Returns the size of the buffer."""
         return len(self.__buf)
@@ -91,7 +145,7 @@ class BinaryReader:
         """
         trimmed = 0
 
-        if size > 0:
+        if size >= 0:
             trimmed = self.size() - size
 
         if (trimmed > 0):
@@ -103,34 +157,32 @@ class BinaryReader:
 
         return trimmed
 
-    def seek(self, offset: int, whence=0) -> None:
+    def seek(self, offset: int, whence: Whence = Whence.BEGIN) -> None:
         """Changes the current position of the buffer by the given offset.\n
         The seek is determined relative to the whence:\n
-        whence = 0 will seek relative to the start.\n
-        whence = 1 will seek relative to the current position.\n
-        whence = 2 will seek relative to the end (offset should be positive).
+        Whence.BEGIN will seek relative to the start.\n
+        Whence.CUR will seek relative to the current position.\n
+        Whence.END will seek relative to the end (offset should be positive).
         """
-        if whence == 0:
-            if offset > len(self.__buf):
-                raise Exception(
-                    'BinaryReader Error: cannot seek farther than buffer length.')
-            else:
-                self.__idx = offset
-        elif whence == 1:
-            if self.__idx + offset > len(self.__buf):
-                raise Exception(
-                    'BinaryReader Error: cannot seek farther than buffer length.')
-            else:
-                self.__idx += offset
-        elif whence == 2:
-            if offset >= len(self.__buf):
-                raise Exception(
-                    'BinaryReader Error: cannot seek farther than buffer length.')
-            else:
-                self.__idx = (len(self.__buf) - 1) - offset
+        new_offset = self.__idx
+
+        if whence == Whence.BEGIN:
+            new_offset = offset
+        elif whence == Whence.CUR:
+            new_offset = self.__idx + offset
+        elif whence == Whence.END:
+            new_offset = len(self.__buf) - offset
+        else:
+            raise Exception('BinaryReader Error: invalid whence value.')
+
+        if self.__past_eof(new_offset) or new_offset < 0:
+            raise Exception(
+                'BinaryReader Error: cannot seek farther than buffer length.')
+
+        self.__idx = new_offset
 
     @contextmanager
-    def seek_to(self, offset: int, whence=0) -> 'BinaryReader':
+    def seek_to(self, offset: int, whence: Whence = Whence.BEGIN) -> 'BinaryReader':
         """Same as `seek(offset, whence)`, but can be used with the `with` statement in a new context.\n
         Upon returning to the old context, the original position of the buffer before the `with` statement will be restored.\n
         Will return a reference of the BinaryReader to be used for `as` in the `with` statement.\n
@@ -142,9 +194,9 @@ class BinaryReader:
 
         self.__idx = prev_pos
 
-    def set_endian(self, is_big_endian: bool) -> None:
+    def set_endian(self, endianness: Endian) -> None:
         """Sets the endianness of the BinaryReader."""
-        self.__big_end = is_big_endian
+        self.__endianness = endianness
 
     def set_encoding(self, encoding: str) -> None:
         """Sets the default encoding of the BinaryReader when reading/writing strings.\n
@@ -153,17 +205,26 @@ class BinaryReader:
         str.encode('', encoding)
         self.__encoding = encoding
 
+    @staticmethod
+    def is_iterable(x) -> bool:
+        return hasattr(x, '__iter__') and not isinstance(x, (str, bytes))
+
     def __read_type(self, format: str, count=1):
         i = self.__idx
-        self.__idx += FMT[format] * count
+        new_offset = self.__idx + (FMT[format] * count)
 
-        end = ">" if self.__big_end else "<"
+        end = ">" if self.__endianness else "<"
 
+        if self.__past_eof(new_offset):
+            raise Exception(
+                'BinaryReader Error: cannot read farther than buffer length.')
+
+        self.__idx = new_offset
         return struct.unpack_from(end + str(count) + format, self.__buf, i)
 
     def read_bytes(self, length=1) -> bytes:
         """Reads a bytes object with the given length from the current position."""
-        return self.__read_type("s", length)
+        return self.__read_type("s", length)[0]
 
     def read_str(self, length=0, encoding=None) -> str:
         """Reads a string with the given length from the current position.\n
@@ -182,7 +243,7 @@ class BinaryReader:
 
             return string.split(b'\x00', 1)[0].decode(encode)
 
-        return self.read_bytes(length)[0].split(b'\x00', 1)[0].decode(encode)
+        return self.read_bytes(length).split(b'\x00', 1)[0].decode(encode)
 
     def read_int64(self, count=1) -> Union[int, Tuple[int]]:
         """Reads a signed 64-bit integer.\n
@@ -264,10 +325,31 @@ class BinaryReader:
             return self.__read_type("e", count)
         return self.__read_type("e")[0]
 
+    def read_struct(self, cls: type, count=1, *args) -> BrStruct:
+        """Creates and returns an instance of the given `cls` after calling its `__br_read__` method.\n
+        `cls` must be a subclass of BrStruct.\n
+        If count is not equal to 1, will return a tuple of values instead of 1 value.\n
+        Additional arguments given after `count` will be passed to the `__br_read__` method of `cls`.\n
+        """
+        if not (cls and issubclass(cls, BrStruct)):
+            raise Exception(
+                f'BinaryReader Error: {cls} is not a subclass of BrStruct.')
+
+        result = []
+        for _ in range(count):
+            br_struct = cls()
+            br_struct.__br_read__(self, *args)
+            result.append(br_struct)
+
+        if count == 1:
+            return result[0]
+
+        return tuple(result)
+
     def __write_type(self, format: str, value, is_iterable: bool) -> None:
         i = self.__idx
 
-        end = ">" if self.__big_end else "<"
+        end = ">" if self.__endianness else "<"
 
         count = 1
         if is_iterable or type(value) is bytes:
@@ -298,62 +380,78 @@ class BinaryReader:
         self.write_bytes(bytes_obj)
         return len(bytes_obj)
 
-    def write_int64(self, value: int, is_iterable=False) -> None:
+    def write_int64(self, value: int) -> None:
         """Writes a signed 64-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("q", value, is_iterable)
+        self.__write_type("q", value, self.is_iterable(value))
 
-    def write_uint64(self, value: int, is_iterable=False) -> None:
+    def write_uint64(self, value: int) -> None:
         """Writes an unsigned 64-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("Q", value, is_iterable)
+        self.__write_type("Q", value, self.is_iterable(value))
 
-    def write_int32(self, value: int, is_iterable=False) -> None:
+    def write_int32(self, value: int) -> None:
         """Writes a signed 32-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("i", value, is_iterable)
+        self.__write_type("i", value, self.is_iterable(value))
 
-    def write_uint32(self, value: int, is_iterable=False) -> None:
+    def write_uint32(self, value: int) -> None:
         """Writes an unsigned 32-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("I", value, is_iterable)
+        self.__write_type("I", value, self.is_iterable(value))
 
-    def write_int16(self, value: int, is_iterable=False) -> None:
+    def write_int16(self, value: int) -> None:
         """Writes a signed 16-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("h", value, is_iterable)
+        self.__write_type("h", value, self.is_iterable(value))
 
-    def write_uint16(self, value: int, is_iterable=False) -> None:
+    def write_uint16(self, value: int) -> None:
         """Writes an unsigned 16-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("H", value, is_iterable)
+        self.__write_type("H", value, self.is_iterable(value))
 
-    def write_int8(self, value: int, is_iterable=False) -> None:
+    def write_int8(self, value: int) -> None:
         """Writes a signed 8-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("b", value, is_iterable)
+        self.__write_type("b", value, self.is_iterable(value))
 
-    def write_uint8(self, value: int, is_iterable=False) -> None:
+    def write_uint8(self, value: int) -> None:
         """Writes an unsigned 8-bit integer.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("B", value, is_iterable)
+        self.__write_type("B", value, self.is_iterable(value))
 
-    def write_float(self, value: float, is_iterable=False) -> None:
+    def write_float(self, value: float) -> None:
         """Writes a 32-bit float.\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("f", value, is_iterable)
+        self.__write_type("f", value, self.is_iterable(value))
 
-    def write_half_float(self, value: float, is_iterable=False) -> None:
+    def write_half_float(self, value: float) -> None:
         """Writes a 16-bit float (half-float).\n
-        If is_iterable is True, will write all of the values in the given iterable.
+        If value is iterable, will write all of the elements in the given iterable.
         """
-        self.__write_type("e", value, is_iterable)
+        self.__write_type("e", value, self.is_iterable(value))
+
+    def write_struct(self, value: BrStruct, *args) -> None:
+        """Calls the given value's `__br_write__` method.\n
+        `value` must be an instance of a class that inherits BrStruct.\n
+        If value is iterable, will call the `__br_write__` method of all elements in the given iterable.\n
+        Additional arguments given after `value` will be passed to the `__br_write__` method of `value`.\n
+        """
+        if not isinstance(value, BrStruct) and not (self.is_iterable(value) and all(isinstance(e, BrStruct) for e in value)):
+            raise Exception(
+                f'BinaryReader Error: {value} is not an instance of BrStruct.')
+
+        if self.is_iterable(value):
+            for s in value:
+                s.__br_write__(self, *args)
+        else:
+            value.__br_write__(self, *args)
